package me.schaertl.graphcli;

import org.eclipse.rdf4j.model.ValueFactory;
import org.eclipse.rdf4j.repository.Repository;
import org.eclipse.rdf4j.repository.RepositoryConnection;
import org.eclipse.rdf4j.repository.RepositoryException;
import org.eclipse.rdf4j.repository.manager.RemoteRepositoryManager;
import org.eclipse.rdf4j.repository.manager.RepositoryManager;

import java.util.Collection;
import java.util.logging.LogManager;

/**
 * Wrapper around various RDF4J methods. Represents a connection to a GraphDB
 * instance on the network.
 *
 * For creating a new connection, simply create a new object of this class. To
 * then close all associated resources, call {@link GraphDB#close}. As this class
 * implements {@link AutoCloseable}, you can also use the convenient try-with-resources.
 */
public class GraphDB implements AutoCloseable {
    private final RepositoryManager manager;
    private final Repository repository;
    private final RepositoryConnection connection;

    private boolean closed = false;

    /**
     * Establish a new connection to a remote GraphDB instance.
     *
     * @param serverURL The URL under which the instance is reachable.
     * @param repository The repository to connect to.
     * @throws RepositoryException If connecting to the server or repository failed.
     */
    public GraphDB(String serverURL, String repository) throws RepositoryException {
        try {
            this.manager = new RemoteRepositoryManager(serverURL);
            this.manager.init();
            final Collection<Repository> rs = this.manager.getAllRepositories();

            this.repository = this.manager.getRepository(repository);
            if (this.repository == null) {
                this.manager.shutDown();
                throw new RepositoryException(String.format("no repository with name %s", repository));
            }

            this.connection = this.repository.getConnection();
        } catch (Exception e) {
            try {
                this.close();
            } catch (Exception ce) {
                // ignore the close exception; we have enough already
            }

            throw e;
        }
    }

    /**
     * Release all associated resources
     *
     * @throws IllegalStateException If this method called multiple times on a single object.
     */
    @Override
    public synchronized void close() throws Exception {
        if (this.closed) {
            throw new IllegalStateException("multiple calls to close()");
        }

        this.closed = true;

        // What comes now is very very ugly. From what I can tell the way
        // the RDF4j API works this is the only way to make sure that everything
        // gets closed properly. Bummer.

        Exception error = null;

        try {
            if (this.connection != null) {
                this.connection.close();
            }
        } catch (Exception e) {
            error = e;
        }

        try {
            if (this.repository != null) {
                this.repository.shutDown();
            }
        } catch (Exception e) {
            if (error != null) {
                error = e;
            }
        }

        try {
            if (this.manager != null) {
                this.manager.shutDown();
            }
        } catch (Exception e) {
            if (error != null) {
                error = e;
            }
        }

        if (error != null) {
            throw error;
        }
    }

    /**
     * Return a connection for interaction with the underlying database.
     *
     * @throws IllegalStateException If {@link GraphDB#close} was called on this object before.
     */
    public synchronized RepositoryConnection connection() {
        if (this.closed) {
            throw new IllegalStateException("called connection() after close()");
        }

        return this.connection;
    }

    /**
     * Return a factory for create new objects.
     *
     * @throws IllegalStateException If {@link GraphDB#close} was called on this object before.
     */
    public synchronized ValueFactory factory() {
        if (this.closed) {
            throw new IllegalStateException("called factory() after close()");
        }

        return this.repository.getValueFactory();
    }
}
