package info.mathhub.uloapi.html;

import org.eclipse.jetty.http.HttpStatus;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import spark.*;
import spark.template.freemarker.FreeMarkerEngine;

import java.util.HashMap;
import java.util.Map;

public class Errors {
    private Errors() {};

    private static final Logger log = LoggerFactory.getLogger(Errors.class);

    /**
     * Singleton instance of the Freemarker engine for use with errors.
     * Used in {@link Errors#createError}.
     */
    private static final FreeMarkerEngine freemarker = new FreeMarkerEngine();

    public static final Route notFound = (Request request, Response response) -> {
        return createError(HttpStatus.NOT_FOUND_404).handle(request, response);
    };

    public static final Route notAcceptable = (Request request, Response response) -> {
        return createError(HttpStatus.NOT_ACCEPTABLE_406).handle(request, response);
    };

    public static final Route notImplemented = (Request request, Response response) -> {
        return createError(HttpStatus.NOT_IMPLEMENTED_501).handle(request, response);
    };

    public static final Route internalServerError = (Request request, Response response) -> {
        return createError(HttpStatus.INTERNAL_SERVER_ERROR_500).handle(request, response);
    };

    public static final ExceptionHandler<Exception> exceptionHandler = (exception, request, response) -> {
        final Map<String, Object> model = new HashMap<>();
        model.put("exception_message", exception.getMessage());

        try {
            createError(HttpStatus.INTERNAL_SERVER_ERROR_500, model).handle(request, response);
        } catch (Exception e) {
            log.error(e.getMessage());
            for (final StackTraceElement element : e.getStackTrace()) {
                log.error(element.toString());
            }
        }
    };

    @SafeVarargs
    private static Route createError(int status, Map<String, Object> ...additionalModels) {
        return (Request request, Response response) -> {
            // set content type

            response.type("text/html");

            // prepare model

            final Map<String, Object> model = new HashMap<>();
            model.put("http_status", status);
            model.put("http_description", HttpStatus.getMessage(status));

            for (final Map<String, Object> additionalModel : additionalModels) {
                model.putAll(additionalModel);
            }

            // render

            final ModelAndView modelAndView = new ModelAndView(model, "error.flt");

            synchronized (Errors.freemarker) {
                return Errors.freemarker.render(modelAndView);
            }
        };
    }
}
