package info.mathhub.uloapi.query;

import org.eclipse.rdf4j.repository.Repository;
import org.eclipse.rdf4j.repository.RepositoryConnection;
import org.eclipse.rdf4j.repository.RepositoryException;
import org.eclipse.rdf4j.repository.manager.RemoteRepositoryManager;
import org.eclipse.rdf4j.repository.manager.RepositoryManager;

public class GraphDB {
    private GraphDB() {};

    static class OperationException extends RuntimeException {
        OperationException(String message) {
            super(message);
        }

        OperationException(String message, Throwable cause) {
            super(message, cause);
        }
    }

    /**
     * Functional interface that defines a function for running operations on
     * a GraphDB database.
     *
     * @param <T> Return type of the operation.
     */
    public interface Operation<T> {
        /**
         * Run some operation connected to a database.
         *
         * @param manager The manager that was used for initializing the database connections.
         * @param repository The repository on which to operate.
         * @param connection The connection to the repository on which to operate.
         * @throws OperationException If the logic of the operation itself causes some kind of error.
         * @throws RepositoryException If communication with the remote database fails.
         */
        T op(RepositoryManager manager, Repository repository, RepositoryConnection connection);
    }

    /**
     * Functional interface that defines a function that takes nothing, does something
     * and returns nothing.
     *
     * Functional programmers hate this.
     */
    private interface Procedure {
        void apply() throws Exception;
    }

    /**
     * Run an operation on a remote repository.
     *
     * @param serverUrl The server of the remote GraphDB instance.
     * @param repositoryName The repository on the remote instance on which to run the operation.
     * @param op The operation to execute.
     * @param <T> The return type of the operation
     * @return The return value of the operation, i.e. of argument {@code op}.
     * @throws OperationException If the logic of the operation itself causes some kind of error.
     * @throws RepositoryException If communication with the remote database fails.
     */
    public static <T> T execute(String serverUrl, String repositoryName, Operation<T> op) {
        RepositoryManager manager = null;
        Repository repository = null;
        RepositoryConnection connection = null;

        Exception err = null;
        T result = null;

        try {
            manager = new RemoteRepositoryManager(serverUrl);
            manager.init();

            if ((repository = manager.getRepository(repositoryName)) == null) {
                throw new RepositoryException("no such repository");
            }

            connection = repository.getConnection();
            result = op.op(manager, repository, connection);
        } catch (Exception e) {
            err = e;
        } finally {
            if (connection != null) {
                err = chain(err, connection::close);
            }

            if (repository != null) {
                err = chain(err, repository::shutDown);
            }

            if (manager != null) {
                err = chain(err, manager::shutDown);
            }
        }

        if (err != null) {
            throw new OperationException("execute failed", err);
        }

        return result;
    }

    /**
     * Execute proc and return {@code e} if not {@code null}, the exception
     * caused by {@code proc} or {@code null}.
     * @return If {@code e} is not {@code null}, always return {@code e}.
     * Otherwise, if {@code proc} caused an exception, return that exception or
     * return {@code null} if {@code proc} did not cause an exception.
     */
    private static Exception chain(Exception e, Procedure proc) {
        Exception ret = e;

        try {
            proc.apply();
        } catch (Exception ne) {
            if (ret == null) {
                ret = ne;
            }
        }

        return ret;
    }
}

