package info.mathhub.uloapi.query;

import info.mathhub.uloapi.config.Config;
import org.eclipse.rdf4j.model.Statement;
import org.eclipse.rdf4j.repository.RepositoryResult;

import java.io.File;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

/**
 * Class {@link Query} wraps various common queries to an RDF4j endpoint. Objects of type {@link Query}
 * are safe for concurrent used by multiple threads.
 */
public class Query {
    private final String serverUrl;
    private final String repository;

    private static Query singleton;

    /**
     * Create a new object for the given endpoint.
     *
     * @param serverUrl URL of the RDF4J endpoint. Usually some kind of HTTP(S) URI.
     * @param repository Name of the repository to connect to.
     */
    public Query(String serverUrl, String repository) {
        this.serverUrl = serverUrl;
        this.repository = repository;
    }

    /**
     * Get {@link Query} instance that is configured to connect to the endpoint defined
     * in {@code uloapi.properties}.
     */
    public static synchronized Query getSingleton() {
        if (Query.singleton == null) {
            Query.singleton = new Query(Config.serverUrl(), Config.repository());
        }

        return Query.singleton;
    }

    /**
     * Return {@link Statistics} that contain some meta data about the connected repository.
     */
    public Statistics getStatistics() {
        final GraphDB.Operation<Statistics> operation = (manager, repository, connection) -> {
            final Statistics statistics = new Statistics();

            statistics.isWriteable = repository.isWritable();
            statistics.numTriplets = connection.size();

            {
                final File dataDir = repository.getDataDir();

                if (dataDir != null) {
                    statistics.dataDir = dataDir.getAbsolutePath();
                }
            }

            return statistics;
        };

        return GraphDB.execute(this.serverUrl, this.repository, operation);
    }

    /**
     * Return a list of triplets that define some kind of {@link ULO#inductive_on} relationship.
     *
     * @return The triplets. Limited to a small number, the actual number of matching triplets in the
     *         underlying RDF4j repository might be way bigger.
     */
    public List<Statement> getInductiveStatements() {
        final GraphDB.Operation<List<Statement>> operation = (manager, repository, connection) -> {
            final RepositoryResult<Statement> result = connection.getStatements(null, ULO.inductive_on, null);
            final Iterator<Statement> iterator = result.iterator();

            final List<Statement> statements = new ArrayList<>();

            for (int i = 0; i < 16 && iterator.hasNext(); i++) {
                final Statement statement = iterator.next();
                statements.add(statement);
            }

            return statements;
        };

        return GraphDB.execute(this.serverUrl, this.repository, operation);
    }
}
